"""This module contains PUS data structures and helpers common for both PUS telemetry and
telecommands.

Content:

* :py:class:`tmtccmd.pus.VerificationWrapper` helper class
"""

from enum import IntEnum
from typing import Optional

from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.pus_1_verification import RequestId
import spacepackets.ecss.pus_1_verification as pus_1
from spacepackets.ecss.pus_verificator import (
    VerificationStatus,
    StatusField,
    PusVerificator,
    TmCheckResult,
)
import logging

from tmtccmd.util.conf_util import AnsiColors


class CustomFsfwPusService(IntEnum):
    SERVICE_200_MODE = 200


class VerificationWrapper:
    def __init__(
        self,
        pus_verificator: PusVerificator,
        console_logger: Optional[logging.Logger],
        file_logger: Optional[logging.Logger],
    ):
        self.pus_verificator = pus_verificator
        self.console_logger = console_logger
        self.file_logger = file_logger
        self.with_colors = True

    @property
    def verificator(self) -> PusVerificator:
        return self.pus_verificator

    def dlog(self, log_str: str, level: int = logging.INFO):
        if self.console_logger is not None:
            self.console_logger.log(level, log_str)
        elif self.file_logger is not None:
            self.file_logger.info(level, log_str)

    def add_tc(self, pus_tc: PusTelecommand) -> bool:
        return self.pus_verificator.add_tc(pus_tc)

    def add_tm(self, srv_1_tm: pus_1.Service1Tm) -> TmCheckResult:
        return self.pus_verificator.add_tm(srv_1_tm)

    def log_to_console(self, srv_1_tm: pus_1.Service1Tm, res: TmCheckResult):
        self.log_to_console_from_req_id(
            srv_1_tm.tc_req_id, res, pus_1.Subservice(srv_1_tm.subservice)
        )

    def log_to_console_from_req_id(
        self,
        req_id: RequestId,
        res: TmCheckResult,
        subservice: Optional[pus_1.Subservice] = None,
    ):
        return self.log_progress_to_console_from_status(res.status, req_id, subservice)

    def log_to_file(self, srv_1_tm: pus_1.Service1Tm, res: TmCheckResult):
        self.log_to_file_from_req_id(
            srv_1_tm.tc_req_id, res, pus_1.Subservice(srv_1_tm.subservice)
        )

    def log_to_file_from_req_id(
        self,
        req_id: RequestId,
        res: TmCheckResult,
        subservice: Optional[pus_1.Subservice] = None,
    ):
        self.log_to_file_from_status(res.status, req_id, subservice)

    def log_to_file_from_status(
        self,
        status: VerificationStatus,
        req_id: RequestId,
        subservice: Optional[pus_1.Subservice] = None,
    ):
        if self.file_logger is None:
            raise ValueError("No valid file logger was set")
        acc_char = gen_file_char_from_status(status.accepted)
        start_char = gen_file_char_from_status(status.started)
        step_char = gen_file_char_from_status(status.step)
        fin_char = gen_file_char_from_status(status.completed)
        step_num = self.step_num(status)
        first_str = self._get_info_string(subservice)
        second_str = f"Request ID {req_id.as_u32():#04x}"
        completion_str = ""
        if status.completed == StatusField.SUCCESS:
            completion_str = " S"
        third_str = (
            f"acc ({acc_char}) sta ({start_char}) ste ({step_char}, {step_num}) "
            f"fin ({fin_char}){completion_str}"
        )

        self.file_logger.info(f"{first_str} | {second_str} | {third_str}")

    def log_progress_to_console_from_status(
        self,
        status: VerificationStatus,
        req_id: RequestId,
        subservice: Optional[pus_1.Subservice] = None,
    ):
        if self.console_logger is None:
            raise ValueError("Invalid console logger")
        acc_char = gen_console_char_from_status(status.accepted, self.with_colors)
        start_char = gen_console_char_from_status(status.started, self.with_colors)
        step_char = gen_console_char_from_status(status.step, self.with_colors)
        fin_char = gen_console_char_from_status(status.completed, self.with_colors)
        step_num = self.step_num(status)
        first_str = self._get_info_string(subservice)
        second_str = f"Request ID {req_id.as_u32():#04x}"
        completion_str = ""
        if status.completed == StatusField.SUCCESS:
            completion_str = f" {AnsiColors.BOLD}{AnsiColors.YELLOW}\u2728"
        third_str = (
            f"acc ({acc_char}) sta ({start_char}) ste ({step_char}, {step_num}) "
            f"fin ({fin_char}){completion_str}"
        )
        self.console_logger.info(f"{first_str} | {second_str} | {third_str}")

    @staticmethod
    def step_num(status: VerificationStatus):
        if not status.step_list:
            return "0"
        else:
            return f"{max(status.step_list)}"

    @staticmethod
    def _get_info_string(subservice: pus_1.Subservice):
        status_str = "Status"
        if subservice is not None:
            if subservice == pus_1.Subservice.TM_ACCEPTANCE_SUCCESS:
                status_str = "Acceptance success"
            elif subservice == pus_1.Subservice.TM_ACCEPTANCE_FAILURE:
                status_str = "Acceptance failure"
            elif subservice == pus_1.Subservice.TM_START_SUCCESS:
                status_str = "Start success"
            elif subservice == pus_1.Subservice.TM_START_FAILURE:
                status_str = "Start failure"
            elif subservice == pus_1.Subservice.TM_STEP_SUCCESS:
                status_str = "Step success"
            elif subservice == pus_1.Subservice.TM_STEP_FAILURE:
                status_str = "Step failure"
            elif subservice == pus_1.Subservice.TM_COMPLETION_SUCCESS:
                status_str = "Completion success"
            elif subservice == pus_1.Subservice.TM_COMPLETION_FAILURE:
                status_str = "Completion failure"
        return f"{status_str} of TC".ljust(25)


def gen_file_char_from_status(status: StatusField):
    if status == StatusField.UNSET:
        return "-"
    elif status == StatusField.FAILURE:
        return "F"
    elif status == StatusField.SUCCESS:
        return "S"


def gen_console_char_from_status(status: StatusField, with_color: bool):
    if status == StatusField.UNSET:
        return dash_unicode(with_color)
    elif status == StatusField.SUCCESS:
        return tick_mark_unicode(with_color)
    elif status == StatusField.FAILURE:
        return cross_mark_unicode(with_color)


def dash_unicode(with_color: bool) -> str:
    if with_color:
        return f"{AnsiColors.YELLOW}-{AnsiColors.RESET}"
    else:
        return "-"


def tick_mark_unicode(with_color: bool) -> str:
    if with_color:
        return f"{AnsiColors.GREEN}\u2713{AnsiColors.RESET}"
    else:
        return "\u2713"


def cross_mark_unicode(with_color: bool) -> str:
    if with_color:
        return f"{AnsiColors.RED}\u274c{AnsiColors.RESET}"
    else:
        return "\u274c"



from threading import Lock
from typing import Optional
import warnings

warnings.warn(
    "the globals_manager module is deprecated", DeprecationWarning, stacklevel=2
)

__GLOBALS_DICT = dict()
__LOCK_TIMEOUT = 50
__GLOBALS_LOCK = Lock()


def get_global(global_param_id: int, lock: bool = False):
    global __LOCK_TIMEOUT, __GLOBALS_DICT
    if lock:
        __GLOBALS_LOCK.acquire(timeout=__LOCK_TIMEOUT)
    global_param = __GLOBALS_DICT.get(global_param_id)
    if lock:
        __GLOBALS_LOCK.release()
    return global_param


def update_global(global_param_id: int, parameter: any, lock: bool = False):
    global __LOCK_TIMEOUT, __GLOBALS_DICT
    if lock:
        __GLOBALS_LOCK.acquire(timeout=__LOCK_TIMEOUT)
    __GLOBALS_DICT[global_param_id] = parameter
    if lock:
        __GLOBALS_LOCK.release()


def lock_global_pool(
    blocking: Optional[bool] = None, timeout: Optional[float] = None
) -> bool:
    """Lock the global objects. This is important if the values are changed. Don't forget to unlock
    the pool after finishing work with the globals!
    :param timeout_seconds: Attempt to lock for this many second. Default value -1 blocks
    permanently until lock is released.
    :return: Returns whether lock was locked or not.
    """
    global __LOCK_TIMEOUT, __GLOBALS_LOCK
    if blocking is None:
        blocking = True
    if timeout is None:
        timeout = __LOCK_TIMEOUT
    return __GLOBALS_LOCK.acquire(blocking=blocking, timeout=timeout)


def unlock_global_pool():
    global __GLOBALS_LOCK
    """Releases the lock so other objects can use the global pool as well"""
    return __GLOBALS_LOCK.release()


def set_lock_timeout(timeout: float):
    global __LOCK_TIMEOUT
    """Set the timeout for the globals manager lock which can ensure thread-safety"""
    __LOCK_TIMEOUT = timeout



from unittest.mock import MagicMock

from tmtccmd.config import HookBase


def create_hook_mock() -> HookBase:
    """Create simple minimal hook mock using the MagicMock facilities by unittest
    :return:
    """
    tmtc_hook_base = MagicMock(spec=HookBase)
    return tmtc_hook_base



import argparse
import os
from unittest import TestCase
from unittest.mock import MagicMock, call, patch

from spacepackets.cfdp import TransmissionMode

from tests.hook_obj_mock import create_hook_mock
from tmtccmd import CoreModeConverter, CoreModeList
from tmtccmd.config import CfdpParams
from tmtccmd.config.args import (
    CommandingParams,
    TreeCommandingParams,
    SetupParams,
    args_to_all_params_tmtc,
    cfdp_args_to_cfdp_params,
    perform_tree_printout,
)
from tmtccmd.config.tmtc import CmdTreeNode, TreePart


class TestArgs(TestCase):
    def setUp(self) -> None:
        self.pargs = argparse.Namespace()
        self.params = SetupParams()
        self.hook_mock = create_hook_mock()

    def base_cli_set(self):
        self.pargs.gui = None
        self.pargs.mode = None
        self.pargs.delay = None
        self.pargs.com_if = "dummy"
        self.pargs.listener = True
        self.pargs.prompt_proc = False
        self.pargs.cmd_path = None
        self.pargs.print_tree = None

    def simple_pargs_cli_set(self):
        self.base_cli_set()
        self.pargs.listener = False
        self.pargs.cmd_path = "/PING"

    def auto_listener_cli_set(self):
        self.base_cli_set()
        self.pargs.listener = True

    def test_basic(self):
        # For some reason, those fields need to be reset manually
        self.params.backend_params.mode = ""
        self.params.backend_params.com_if_id = ""
        self.simple_pargs_cli_set()
        self.assertEqual(self.params.cmd_params.delay, 0)
        self.assertEqual(self.params.backend_params.mode, "")
        self.assertEqual(self.params.backend_params.com_if_id, "")
        def_params = TreeCommandingParams(None)
        args_to_all_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        # Set to default value
        self.assertEqual(self.params.cmd_params.delay, 4.0)
        # Unset
        self.assertEqual(self.params.cmd_params.apid, 0)
        self.assertEqual(self.params.app_params.use_gui, False)
        self.assertEqual(self.params.app_params.use_ansi_colors, True)
        self.assertEqual(def_params.cmd_path, "/PING")
        self.assertEqual(
            self.params.backend_params.mode,
            CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE),
        )
        self.assertEqual(self.params.backend_params.listener, False)
        self.assertEqual(self.params.backend_params.com_if_id, "dummy")

    def test_delay_set(self):
        self.simple_pargs_cli_set()
        self.pargs.delay = 2.0
        def_params = TreeCommandingParams(None)
        args_to_all_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        self.assertEqual(def_params.cmd_path, "/PING")
        self.assertEqual(self.params.cmd_params.delay, 2.0)

    def test_cfdp_conversion_basic(self):
        self.pargs.source = "hello.txt"
        self.pargs.target = "hello-dest.txt"
        self.pargs.no_closure = False
        self.pargs.proxy = True
        self.pargs.type = "nak"
        cfdp_params = CfdpParams()
        cfdp_params.transmission_mode = TransmissionMode.ACKNOWLEDGED
        self.assertEqual(cfdp_params.closure_requested, False)
        self.assertEqual(cfdp_params.proxy_op, False)
        cfdp_args_to_cfdp_params(self.pargs, cfdp_params)
        self.assertEqual(cfdp_params.source_file, "hello.txt")
        self.assertEqual(cfdp_params.dest_file, "hello-dest.txt")
        self.assertEqual(cfdp_params.closure_requested, True)
        self.assertEqual(cfdp_params.transmission_mode, TransmissionMode.UNACKNOWLEDGED)
        self.assertEqual(cfdp_params.proxy_op, True)

    def test_cfdp_conversion_acked(self):
        self.pargs.type = "ack"
        cfdp_params = CfdpParams()
        cfdp_args_to_cfdp_params(self.pargs, cfdp_params)
        self.assertEqual(cfdp_params.transmission_mode, TransmissionMode.ACKNOWLEDGED)

    def test_auto_listener_mode(self):
        self.auto_listener_cli_set()
        def_params = TreeCommandingParams(None)
        args_to_all_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        self.assertEqual(self.params.backend_params.listener, True)
        self.assertIsNone(def_params.cmd_path)
        self.assertEqual(
            self.params.backend_params.mode,
            CoreModeConverter.get_str(CoreModeList.LISTENER_MODE),
        )
        self.assertEqual(
            self.params.mode, CoreModeConverter.get_str(CoreModeList.LISTENER_MODE)
        )
        self.params = SetupParams()

    def test_tree_printout_conversion_default(self):
        self.base_cli_set()
        self.pargs.print_tree = []
        def_params = TreeCommandingParams(None)
        args_to_all_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        self.assertTrue(self.params.cmd_params.print_tree)
        self.assertTrue(self.params.cmd_params.tree_print_with_description)
        self.assertIsNone(self.params.cmd_params.tree_print_max_depth)

    def test_tree_printout_conversion_with_custom_args(self):
        self.base_cli_set()
        self.pargs.print_tree = ["b", "2"]
        def_params = TreeCommandingParams(None)
        args_to_all_params_tmtc(
            pargs=self.pargs,
            params=self.params,
            hook_obj=self.hook_mock,
            use_prompts=False,
            def_tmtc_params=def_params,
            assign_com_if=False,
        )
        self.assertTrue(self.params.cmd_params.print_tree)
        self.assertFalse(self.params.cmd_params.tree_print_with_description)
        self.assertEqual(self.params.cmd_params.tree_print_max_depth, 2)

    @patch("builtins.print")
    def test_tree_printout_0(self, print_mock: MagicMock):
        root_node_only = CmdTreeNode.root_node()
        tc_params = CommandingParams()
        tc_params.print_tree = True
        perform_tree_printout(tc_params, root_node_only)
        self.assertEqual(len(print_mock.call_args_list), 2)
        self.assertEqual(
            print_mock.call_args_list[0],
            call("Printing command tree with full descriptions:"),
        )
        printout = print_mock.call_args_list[1].args[0]
        self.assertTrue("/" in printout)
        self.assertTrue("[ Root Node ]" in printout)

    @patch("builtins.print")
    def test_tree_printout_1(self, print_mock: MagicMock):
        root_node_only = CmdTreeNode.root_node()
        tc_params = CommandingParams()
        tc_params.print_tree = True
        tc_params.tree_print_with_description = False
        perform_tree_printout(tc_params, root_node_only)
        self.assertEqual(len(print_mock.call_args_list), 2)
        self.assertEqual(
            print_mock.call_args_list[0],
            call("Printing command tree without descriptions:"),
        )
        self.assertEqual(print_mock.call_args_list[1], call(f"/{os.linesep}"))

    @patch("builtins.print")
    def test_tree_printout_2(self, print_mock: MagicMock):
        root_node_only = CmdTreeNode.root_node()
        root_node_only.add_child(CmdTreeNode("acs", "ACS Subsystem"))
        tc_params = CommandingParams()
        tc_params.print_tree = True
        tc_params.tree_print_with_description = False
        tc_params.tree_print_max_depth = 0
        perform_tree_printout(tc_params, root_node_only)
        self.assertEqual(len(print_mock.call_args_list), 2)
        self.assertEqual(
            print_mock.call_args_list[0],
            call("Printing command tree without descriptions and maximum depth 0:"),
        )
        self.assertEqual(
            print_mock.call_args_list[1],
            call(
                f"/{os.linesep}{TreePart.CORNER.value} ... (cut-off, maximum depth 0){os.linesep}"
            ),
        )



from unittest import TestCase

from prompt_toolkit.history import InMemoryHistory
from tmtccmd.config.prompt import prompt_cmd_path
from tmtccmd.config.tmtc import CmdTreeNode
from unittest.mock import MagicMock, patch


class TestPromptFunc(TestCase):
    def setUp(self) -> None:
        self.cmd_tree = CmdTreeNode.root_node()
        return super().setUp()

    def base_tree(self):
        self.cmd_tree.add_child(CmdTreeNode("acs", "ACS Subsystem"))
        self.cmd_tree.add_child(CmdTreeNode("tcs", "TCS Subsystem"))
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))

    @patch("builtins.print")
    def test_prompt_cmd_path_simple_print(self, _: MagicMock):
        with patch(
            "tmtccmd.config.prompt.prompt_toolkit.prompt", side_effect=[":p", "/"]
        ) as prompt:
            cmd_path = prompt_cmd_path(self.cmd_tree)
            prompt.assert_not_called()
            self.assertEqual(cmd_path, "/")

    @patch("builtins.print")
    def test_prompt_cmd_path_full_print(self, mocked_print: MagicMock):
        self.base_tree()
        with patch(
            "tmtccmd.config.prompt.prompt_toolkit.prompt", side_effect=[":pf", "acs"]
        ) as prompt:
            cmd_path = prompt_cmd_path(self.cmd_tree)
            self.assertEqual(len(prompt.call_args_list), 2)
            self.assertEqual(cmd_path, "/acs")
            call_list = mocked_print.call_args_list
            self.assertEqual(len(call_list), 2)
            help_txt = mocked_print.call_args_list[0].args[0]
            self.assertTrue("Additional commands for prompt:" in help_txt)
            self.assertTrue(":p[b][f][<depth>] Tree Print" in help_txt)
            self.assertTrue(":r Retry" in help_txt)
            self.assertTrue(":h Help Text" in help_txt)
            printout = mocked_print.call_args_list[1].args[0]
            self.assertTrue("acs" in printout)
            self.assertTrue("ACS Subsystem" in printout)
            self.assertTrue("tcs" in printout)
            self.assertTrue("TCS Subsystem" in printout)

    def test_prompt_cmd_path_retry_explicit(self):
        self.base_tree()
        with patch(
            "tmtccmd.config.prompt.prompt_toolkit.prompt",
            side_effect=["acs/:r", "ping"],
        ) as prompt:
            cmd_path = prompt_cmd_path(self.cmd_tree)
            self.assertEqual(cmd_path, "/ping")
            self.assertEqual(len(prompt.call_args_list), 2)

    def test_prompt_cmd_path_retry_prompted(self):
        self.base_tree()
        with patch(
            "tmtccmd.config.prompt.prompt_toolkit.prompt",
            side_effect=["acss", "acs"],
        ) as prompt_mock:
            with patch(
                "tmtccmd.config.prompt.input",
                return_value="yes",
            ) as input_mock:
                cmd_path = prompt_cmd_path(self.cmd_tree)
                self.assertEqual(cmd_path, "/acs")
                self.assertEqual(len(prompt_mock.call_args_list), 2)
                self.assertEqual(len(input_mock.call_args_list), 1)

    @patch("builtins.print")
    def test_prompt_help_reprint(self, mocked_print: MagicMock):
        self.base_tree()
        with patch(
            "tmtccmd.config.prompt.prompt_toolkit.prompt", side_effect=[":h", "acs"]
        ) as prompt:
            cmd_path = prompt_cmd_path(self.cmd_tree)
            self.assertEqual(len(prompt.call_args_list), 2)
            self.assertEqual(cmd_path, "/acs")
            call_list = mocked_print.call_args_list
            self.assertEqual(len(call_list), 2)
            help_txt = mocked_print.call_args_list[0].args[0]
            self.assertTrue("Additional commands for prompt:" in help_txt)
            self.assertTrue(":p[b][f][<depth>] Tree Print" in help_txt)
            self.assertTrue(":r Retry" in help_txt)
            self.assertTrue(":h Help Text" in help_txt)
            help_txt_2 = mocked_print.call_args_list[1].args[0]
            self.assertEqual(help_txt, help_txt_2)

    def test_cmd_history(self):
        self.base_tree()
        history = InMemoryHistory()
        history.append_string("test")
        with patch(
            "tmtccmd.config.prompt.prompt_toolkit.prompt", side_effect=["ping"]
        ) as prompt:
            cmd_path = prompt_cmd_path(self.cmd_tree, history=history)
            self.assertEqual(len(prompt.call_args_list), 1)
            self.assertEqual(cmd_path, "/ping")



import enum
import logging
from abc import abstractmethod, ABC
from typing import Any, Deque, List, Union, Dict, Optional

from spacepackets.ecss import PusTelemetry

from .queue import (
    QueueHelperBase,
    DefaultPusQueueHelper,
    QueueWrapper,
    TcQueueEntryType,
    TcQueueEntryBase,
    QueueEntryHelper,
    WaitEntry,
    SpacePacketEntry,
    PusTcEntry,
    RawTcEntry,
    PacketDelayEntry,
    LogQueueEntry,
)
from .procedure import (
    TcProcedureBase,
    TcProcedureType,
    TreeCommandingProcedure,
    CustomProcedureInfo,
    ProcedureWrapper,
)

from .handler import FeedWrapper, TcHandlerBase, SendCbParams
from .decorator import service_provider, route_to_registered_service_handlers
from .common import *  # noqa re-export
from .ccsds_tm_listener import CcsdsTmListener  # noqa re-export



import atexit
import logging
import sys
from collections import deque
from datetime import timedelta
from typing import Optional

from tmtccmd.core.backend_base import BackendBase
from tmtccmd.core.backend_state import BackendState
from tmtccmd.core.base import TcMode, TmMode, BackendRequest
from tmtccmd.tmtc import TcProcedureBase, ProcedureWrapper
from tmtccmd.tmtc.handler import TcHandlerBase, FeedWrapper
from tmtccmd.util.exit import keyboard_interrupt_handler
from tmtccmd.tmtc.queue import QueueWrapper
from tmtccmd.tmtc.ccsds_seq_sender import (
    SequentialCcsdsSender,
    SenderMode,
)
from tmtccmd.tmtc.ccsds_tm_listener import CcsdsTmListener
from tmtccmd.com import ComInterface


class NoValidProcedureSet(Exception):
    pass


class CcsdsTmtcBackend(BackendBase):
    """This is the primary class which handles TMTC reception and sending"""

    def __init__(
        self,
        tc_mode: TcMode,
        tm_mode: TmMode,
        com_if: ComInterface,
        tm_listener: CcsdsTmListener,
        tc_handler: TcHandlerBase,
    ):
        self._state = BackendState()
        self._state.mode_wrapper.tc_mode = tc_mode
        self._state.mode_wrapper.tm_mode = tm_mode

        self._com_if_active = False
        self._tc_handler = tc_handler

        self._com_if = com_if
        self._tm_listener = tm_listener
        self.exit_on_com_if_init_failure = True
        # This can be used to keep the TC mode in multi queue mode after finishing the handling
        # of a queue
        self.keep_multi_queue_mode = False
        self.keep_listener_mode = False
        self._queue_wrapper = QueueWrapper(None, deque())
        self._seq_handler = SequentialCcsdsSender(
            tc_handler=tc_handler,
            queue_wrapper=self._queue_wrapper,
        )

    def register_keyboard_interrupt_handler(self):
        """Register a keyboard interrupt handler which closes the COM interface and prints
        a small message"""
        atexit.register(keyboard_interrupt_handler, self)

    @property
    def com_if_id(self):
        return self._com_if.id

    @property
    def com_if(self) -> ComInterface:
        return self._com_if

    @property
    def state(self):
        return self._state

    @property
    def request(self):
        return self._state.request

    @property
    def tc_mode(self):
        return self._state.mode_wrapper.tc_mode

    @property
    def tm_mode(self):
        return self._state.mode_wrapper.tm_mode

    @property
    def inter_cmd_delay(self):
        return self._queue_wrapper.inter_cmd_delay

    @inter_cmd_delay.setter
    def inter_cmd_delay(self, delay: timedelta):
        self._queue_wrapper.inter_cmd_delay = delay

    @tc_mode.setter
    def tc_mode(self, tc_mode: TcMode):
        self._state.mode_wrapper.tc_mode = tc_mode

    @tm_mode.setter
    def tm_mode(self, tm_mode: TmMode):
        self._state.mode_wrapper.tm_mode = tm_mode

    @property
    def tm_listener(self):
        return self._tm_listener

    def try_set_com_if(self, com_if: ComInterface) -> bool:
        if not self.com_if_active():
            self._com_if = com_if
            return True
        else:
            return False

    def com_if_active(self):
        return self._com_if_active

    @property
    def current_procedure(self) -> ProcedureWrapper:
        return ProcedureWrapper(self._queue_wrapper.info)

    @current_procedure.setter
    def current_procedure(self, proc_info: TcProcedureBase):
        self._queue_wrapper.info = proc_info

    def start(self):
        self.open_com_if()

    def __listener_io_error_handler(self, ctx: str):
        logger = logging.getLogger(__name__)
        logger.error(f"Communication Interface could not be {ctx}")
        logger.info("TM listener will not be started")
        if self.exit_on_com_if_init_failure:
            logger.error("Closing TMTC commander..")
            self._com_if.close()
            sys.exit(1)

    def open_com_if(self):
        try:
            self._com_if.open()
        except IOError:
            self.__listener_io_error_handler("opened")
        self._com_if_active = True

    def close_com_if(self):
        """Closes the TM listener and the communication interface
        :return:
        """
        try:
            self._com_if.close()
        except IOError:
            self.__listener_io_error_handler("close")
        self._com_if_active = False

    def periodic_op(self, _args: Optional[any] = None) -> BackendState:
        """Periodic operation. Simply calls the :py:meth:`default_operation` function.
        :raises KeyboardInterrupt: Yields info output and then propagates the exception
        :raises IOError: Yields informative output and propagates exception
        :"""
        self.default_operation()
        return self._state

    def default_operation(self):
        """Command handling. This is a convenience function to call the TM and the TC operation
        and then auto-determine the internal mode with the :py:meth:`mode_to_req` method.

        :raises NoValidProcedureSet: No valid procedure set to be passed to the feed callback of
            the TC handler
        """
        self.tm_operation()
        self.tc_operation()
        self.mode_to_req()

    def mode_to_req(self):
        """This function will convert the internal state of the backend to a backend
        :py:attr:`request`,  which can be used to determine the next operation. These requests can
        be treated like recommendations.
        For example, for if both the TC and the TM mode are IDLE, the request will be set to
        :py:attr:`BackendRequest.DELAY_IDLE` field.
        """
        if self.tc_mode == TcMode.IDLE and self.tm_mode == TmMode.IDLE:
            self._state._req = BackendRequest.DELAY_IDLE
        elif self.tm_mode == TmMode.LISTENER and self.tc_mode == TcMode.IDLE:
            self._state._req = BackendRequest.DELAY_LISTENER
        elif self._seq_handler.mode == SenderMode.DONE:
            if self._state.tc_mode == TcMode.ONE_QUEUE:
                if self.keep_listener_mode:
                    self._state._req = BackendRequest.DELAY_LISTENER
                    self.tm_mode = TmMode.LISTENER
                    self.tc_mode = TcMode.IDLE
                else:
                    self.tc_mode = TcMode.IDLE
                    self._state._req = BackendRequest.TERMINATION_NO_ERROR
            elif self._state.tc_mode == TcMode.MULTI_QUEUE:
                if not self.keep_multi_queue_mode:
                    self._state.mode_wrapper.tc_mode = TcMode.IDLE
                self._state._req = BackendRequest.CALL_NEXT
        else:
            if (
                not self._state.sender_res.next_entry_is_tc
                and not self._state.sender_res.queue_empty
            ):
                self._state._req = BackendRequest.CALL_NEXT
            else:
                if (
                    int(self._state.sender_res.longest_rem_delay.microseconds / 1000.0)
                    > 0
                ):
                    self._state._recommended_delay = (
                        self._state.sender_res.longest_rem_delay
                    )
                    self._state._req = BackendRequest.DELAY_CUSTOM
                else:
                    self._state._req = BackendRequest.CALL_NEXT

    def poll_tm(self):
        """Poll TM, irrespective of current TM mode"""
        self._tm_listener.operation(self._com_if)

    def tm_operation(self):
        """This function will fetch and forward TM data from the current communication interface
        to the user TM handler. It only does so if the :py:attr:`tm_mode` is set to the LISTENER
        mode
        """
        if self._state.tm_mode == TmMode.LISTENER:
            self._tm_listener.operation(self._com_if)

    def tc_operation(self):
        """This function will handle consuming the current TC queue
        if one is available, or attempting to fetch a new one if it is not. This function will only
        do something if the :py:attr:`tc_mode` is set to a non IDLE value.

        It is necessary to set a valid procedure before calling this by using the
        :py:attr:`current_proc_info` setter function.

        :raises NoValidProcedureSet: No valid procedure set to be passed to the feed callback of
            the TC handler
        """
        if self._state.tc_mode != TcMode.IDLE:
            self.__check_and_execute_queue()

    def __check_and_execute_queue(self):
        if self._seq_handler.mode == SenderMode.DONE:
            queue = self.__prepare_tc_queue()
            if queue is None:
                return
            logging.getLogger(__name__).info("Loading TC queue")
            self._seq_handler.queue_wrapper = queue
            self._seq_handler.resume()
        self._state._sender_res = self._seq_handler.operation(self._com_if)

    def __prepare_tc_queue(self, auto_dispatch: bool = True) -> Optional[QueueWrapper]:
        feed_wrapper = FeedWrapper(self._queue_wrapper, auto_dispatch)
        if self._queue_wrapper.info is None:
            raise NoValidProcedureSet(
                "No procedure was set to pass to the feed callback function"
            )
        self._tc_handler.feed_cb(
            ProcedureWrapper(self._queue_wrapper.info), feed_wrapper
        )
        if not feed_wrapper.dispatch_next_queue:
            return None
        return feed_wrapper.queue_wrapper



import logging
from typing import Callable, Optional

from PyQt6.QtCore import QThreadPool, QRunnable
from PyQt6.QtWidgets import QPushButton

from tmtccmd import HookBase
from tmtccmd.gui.defs import (
    SharedArgs,
    LocalArgs,
    WorkerOperationsCode,
    DISCONNECT_BTTN_STYLE,
    CONNECT_BTTN_STYLE,
    COMMAND_BUTTON_STYLE,
)
from tmtccmd.gui.defs import FrontendState
from tmtccmd.gui.worker import FrontendWorker

LOGGER = logging.getLogger(__name__)


class ButtonArgs:
    def __init__(
        self,
        state: FrontendState,
        pool: QThreadPool,
        shared: SharedArgs,
    ):
        self.state = state
        self.pool = pool
        self.shared = shared


class ConnectButtonParams:
    def __init__(
        self,
        hook_obj: HookBase,
        connect_cb: Callable[[], None],
        disconnect_cb: Callable[[], None],
        tm_listener_bttn: Optional[QPushButton],
    ):
        self.hook_obj = hook_obj
        self.connect_cb = connect_cb
        self.disconnect_cb = disconnect_cb
        self.tm_listener_bttn = tm_listener_bttn


class ConnectButtonWrapper:
    def __init__(
        self, button: QPushButton, args: ButtonArgs, bttn_params: ConnectButtonParams
    ):
        self.button = button
        self._args = args
        self._bttn_params = bttn_params
        self._connected = False
        self._next_con_state = False
        self._com_if_needs_switch = False
        self.button.clicked.connect(self._button_op)

    def _button_op(self):
        if not self._connected:
            self._connect_button_pressed()
        else:
            self._disconnect_button_pressed()

    def _connect_button_pressed(self):
        LOGGER.info("Opening COM Interface")
        self._com_if_needs_switch = False
        # Build and assign new communication interface
        if self._args.state.current_com_if != self._args.state.last_com_if:
            self._com_if_needs_switch = True
        self.button.setEnabled(False)
        worker = FrontendWorker.spawn_for_opening_com_if(
            self._com_if_needs_switch,
            self._args.state.current_com_if,
            self._bttn_params.hook_obj,
            self._args.shared,
        )
        self._next_con_state = True
        worker.signals.finished.connect(self._button_op_done)
        # TODO: Connect failure signal as well
        self._args.pool.start(worker)

    def _button_op_done(self):
        if self._next_con_state:
            if self._com_if_needs_switch:
                self._args.state.last_com_if = self._args.state.current_com_if
            self._connect_button_finished()
        else:
            self._disconnect_button_finished()
        self._connected = self._next_con_state

    def _connect_button_finished(self):
        self.button.setStyleSheet(DISCONNECT_BTTN_STYLE)
        self.button.setText("Disconnect")
        self.button.setEnabled(True)
        self._bttn_params.connect_cb()
        if (
            self._args.state.auto_connect_tm_listener
            and self._bttn_params.tm_listener_bttn is not None
        ):
            self._bttn_params.tm_listener_bttn.click()
        LOGGER.info("Connected")

    def _disconnect_button_pressed(self):
        self.button.setEnabled(False)
        self._next_con_state = False
        worker = FrontendWorker(
            LocalArgs(WorkerOperationsCode.CLOSE_COM_IF, None), self._args.shared
        )
        worker.signals.finished.connect(self._button_op_done)
        self._args.pool.start(worker)

    def _disconnect_button_finished(self):
        self.button.setEnabled(True)
        self.button.setStyleSheet(CONNECT_BTTN_STYLE)
        self.button.setText("Connect")
        self._bttn_params.disconnect_cb()
        LOGGER.info("Disconnected")


class TmButtonWrapper:
    def __init__(self, button: QPushButton, args: ButtonArgs, conn_button: QPushButton):
        self.button = button
        self.args = args
        self.worker: Optional[QRunnable] = None
        self._listening = False
        self._next_listener_state = False
        self.button.setStyleSheet(CONNECT_BTTN_STYLE)
        self.button.setText("Start TM listener")
        self.button.setEnabled(False)
        self.button.clicked.connect(self.button_op)
        self._conn_button = conn_button

    def is_listening(self):
        return self._listening

    def stop_thread(self):
        if self.worker:
            self.stop_listener()

    def abort_thread(self):
        if self.worker:
            self.worker.signals.abort.emit(None)

    def start_listener(self):
        LOGGER.info("Starting TM listener")
        self.worker = FrontendWorker(
            LocalArgs(WorkerOperationsCode.LISTEN_FOR_TM, 0.4), self.args.shared
        )
        self._next_listener_state = True
        self._conn_button.setDisabled(True)
        self.args.pool.start(self.worker)
        self.button_op_done()

    def stop_listener(self):
        LOGGER.info("Stopping TM listener")
        self._next_listener_state = False
        if self.worker is not None:
            self.worker.signals.finished.connect(self.button_op_done)
            self.worker.signals.stop.emit(None)
        self.button.setEnabled(False)

    def button_op(self):
        if not self._listening:
            self.start_listener()
        else:
            self.stop_listener()

    def button_op_done(self):
        if self._next_listener_state:
            self.button.setStyleSheet(DISCONNECT_BTTN_STYLE)
            self.button.setText("Stop TM listener")
            self._listening = True
            self.button.setEnabled(True)
        else:
            self.button.setStyleSheet(CONNECT_BTTN_STYLE)
            if not self.args.shared.com_if_ref_tracker.is_used():
                self._conn_button.setEnabled(True)
            self.button.setText("Start TM listener")
            self._listening = False
        self.button.setEnabled(True)


class SendButtonWrapper:
    def __init__(self, button: QPushButton, args: ButtonArgs, conn_button: QPushButton):
        self.button = button
        self._args = args
        self._conn_button = conn_button
        self.debug_mode = False
        self.button.setText("Send Command")
        self.button.setStyleSheet(COMMAND_BUTTON_STYLE)
        self.button.setEnabled(False)
        self.button.clicked.connect(self._button_op)

    def _button_op(self):
        if self.debug_mode:
            LOGGER.info("Send command button pressed.")
        self.button.setDisabled(True)
        if self._args.state.current_cmd_path is None:
            return
        worker = FrontendWorker.spawn_for_cmd_path(
            self._args.state.current_cmd_path, self._args.shared
        )
        worker.signals.finished.connect(self._finish_op)
        self._args.pool.start(worker)

    def _finish_op(self):
        self.button.setEnabled(True)
        if not self._args.shared.com_if_ref_tracker.is_used():
            self._conn_button.setEnabled(True)
